package ui;

import java.util.ArrayList;
import java.util.List;

import ui.internal.MDPInterface;
import ui.internal.Maze;
import ui.internal.State;
import ui.internal.StateInterface;

/**
 * 
 * This class represents MDP model. Students are not using this class directly.
 * States are represented by grid cells and by one 'dummy state' outside the
 * grid.
 * 
 * There are following types of states: 'normal state', 'teleport state',
 * 'absorbing state' and 'dummy state'.
 * 
 * All states excluding 'dummy state' are represented by a grid cell.
 * 
 * Next state after 'normal state' can be only state which is its cell neighbor.
 * I.e. to another states is zero probability. If there is no wall around the
 * source state, Pcoorect is probability that the next state will be changed
 * according to used action. In case that there are no walls around a source
 * state, there is zero probability that state will be changed into the opposite
 * direction as was selected by the action. Rest of probability is uniformly
 * divided into remaining two directions.
 * 
 * If sate is surrounded by walls, probability in direction of walls is then
 * zero. Original probability of this directions are summarized and used for
 * staying in the same state probability.
 * 
 * Next state after 'teleport state' is randomly selected state (including this
 * teleport state and excluding the 'dummy state') independently on performed
 * action.
 * 
 * Next state after 'absorbing state' is the 'dummy state' independently on
 * performed action.
 * 
 * Next state after 'dummy state' is the 'dummy state' independently on
 * performed action.
 * 
 * It is possible to define reward for each state using editor excluding the
 * 'dummy state' which is not displayed. Reward of the 'dummy state' is zero.
 * 
 * @author Petr
 */

public class MDPModel implements MDPInterface {

	private Maze maze;

	private double Pcorrect = 0.8;

	// normal include absorbing and teleport state without dummies!
	private int normalStates;

	private State dummyState = new State();

	private List<StudentStateInterface> allStates;

	public MDPModel(Maze maze) {
		this.maze = maze;

		int height = maze.getHeight();
		int width = maze.getWidth();

		allStates = new ArrayList<StudentStateInterface>();

		for (int x = 0; x < width; x++) {
			for (int y = 0; y < height; y++) {
				boolean absorbing = maze.isAbsorbingGoal(x, y);
				boolean teleport = maze.isTeleportState(x, y);

				if (absorbing && teleport) {
					System.out
							.println("State cannot be a teleport ant absorbing simultaneously!");
				}

				double reward = maze.getReward(x, y);

				State s = new State(x, y, reward, absorbing, teleport);

				allStates.add(s);
			}
		}

		this.normalStates = allStates.size();

		allStates.add(dummyState);
	}

	public void setPcorrect(double Pcorrect) {
		this.Pcorrect = Pcorrect;
	}

	public List<StudentStateInterface> getAllStates() {
		return allStates;
	}

	@Override
	public double getTransitionProbability(StudentStateInterface currentState,
			Action action, StudentStateInterface futureState) {

		StateInterface fromState = (StateInterface) currentState;
		StateInterface toState = (StateInterface) futureState;
		
		// dummy state
		if ((fromState == dummyState) && (toState != dummyState)) {
			return 0;
		}

		if ((fromState == dummyState) && (toState == dummyState)) {
			return 1;
		}

		// absorbing state
		if (fromState.isAbsorbingState()) {
			if (futureState != dummyState) {
				return 0;
			}

			if (toState == dummyState) {
				return 1;
			}
		}

		// teleport state

		if (fromState.isTeleportState()) {
			if (toState != dummyState) {
				return 1.0 / normalStates;
			} else {
				return 0;
			}
		}

		// non teleport and non absorbing and non dummy states

		if (toState == dummyState) {
			return 0;
		} else {
			// fromState and toState are not dummies
			int fromX = fromState.getX();
			int fromY = fromState.getY();

			int toX = toState.getX();
			int toY = toState.getY();

			int deltaX = Math.abs(fromX - toX);
			int deltaY = Math.abs(fromY - toY);
			int delta = deltaX + deltaY;

			if (delta > 1) {
				return 0;
			}

			// states are neighbors

			boolean isWallUp;
			boolean isWallDown;
			boolean isWallLeft;
			boolean isWallRight;

			isWallLeft = maze.isWall(fromX, fromY, Action.GO_WEST);
			isWallRight = maze.isWall(fromX, fromY, Action.GO_EAST);
			isWallUp = maze.isWall(fromX, fromY, Action.GO_NORTH);
			isWallDown = maze.isWall(fromX, fromY, Action.GO_SOUTH);

			double otherP = (1 - Pcorrect) / 2;

			double pLeft = 0;
			double pRight = 0;
			double pUp = 0;
			double pDown = 0;
			double pStay = 0;

			if (action.GO_WEST.equals(action)) {
				pLeft = Pcorrect;
				pRight = 0;
				pUp = otherP;
				pDown = otherP;
			}
			if (action.GO_EAST.equals(action)) {
				pLeft = 0;
				pRight = Pcorrect;
				pUp = otherP;
				pDown = otherP;
			}

			if (action.GO_NORTH.equals(action)) {
				pLeft = otherP;
				pRight = otherP;
				pUp = Pcorrect;
				pDown = 0;
			}

			if (action.GO_SOUTH.equals(action)) {
				pLeft = otherP;
				pRight = otherP;
				pUp = 0;
				pDown = Pcorrect;
			}

			if (isWallLeft) {
				pStay += pLeft;
				pLeft = 0;
			}
			if (isWallRight) {
				pStay += pRight;
				pRight = 0;
			}
			if (isWallUp) {
				pStay += pUp;
				pUp = 0;
			}
			if (isWallDown) {
				pStay += pDown;
				pDown = 0;
			}

			//if (delta != 0) {
				
				if ((fromX == toX) && (fromY == toY-1)) {
					return pDown;
				}
				
				if ((fromX == toX) && (fromY == toY+1)) {
					return pUp;
				}
				
				if ((fromX == toX+1) && (fromY == toY)) {
					return pLeft;
				}
				

				if ((fromX == toX-1) && (fromY == toY)) {
					return pRight;
				}
				
				
				if ((fromX == toX) && (fromY == toY)) {
					return pStay;
				}
				
				
				
/*				if (action.GO_LEFT.equals(action)) {
					return pLeft;
				}
				if (action.GO_RIGHT.equals(action)) {
					return pRight;
				}

				if (action.GO_UP.equals(action)) {
					return pUp;
				}

				if (action.GO_DOWN.equals(action)) {
					return pDown;
				} */
				
//			} else {
//				return pStay;
//			}

			System.out.println("Bad action");
			return 0;

		}
	}

	public double getReward(StudentStateInterface state) {
		StateInterface s = (StateInterface) state;
		return s.getReward();
	}

}
